"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.mergeUnionTypes = exports.unwrapMaybePromise = exports.extractContextTypes = void 0;
const tslib_1 = require("tslib");
const Either_1 = require("fp-ts/lib/Either");
const lodash_1 = tslib_1.__importDefault(require("lodash"));
const Path = tslib_1.__importStar(require("path"));
const tsm = tslib_1.__importStar(require("ts-morph"));
const typescript_1 = tslib_1.__importDefault(require("typescript"));
const nexus_logger_1 = require("../nexus-logger");
const utils_1 = require("../utils");
const errors_1 = require("./errors");
const log = nexus_logger_1.rootLogger.child('addToContextExtractor');
function contribTypeRef(name) {
    return { kind: 'ref', name };
}
function contribTypeLiteral(value) {
    return { kind: 'literal', value };
}
/**
 * Extract types from all `addToContext` calls.
 */
function extractContextTypes(program) {
    const typeImportsIndex = {};
    const checker = program.getTypeChecker();
    const contextTypeContributions = {
        typeImports: [],
        types: [],
    };
    const appSourceFiles = program.getSourceFiles().filter((sf) => !sf.fileName.match(/node_modules/));
    log.trace('got app source files', {
        count: appSourceFiles.length,
    });
    const wrappedNodes = appSourceFiles.map((n) => tsm.createWrappedNode(n, { typeChecker: checker }));
    for (const wrappedNode of wrappedNodes) {
        try {
            visit(wrappedNode);
        }
        catch (err) {
            return Either_1.left(err);
        }
    }
    log.trace('finished compiler extension processing', {
        contextTypeContributions,
    });
    // flush deduped type imports
    contextTypeContributions.typeImports.push(...Object.values(typeImportsIndex));
    return Either_1.right(contextTypeContributions);
    /**
     * Given a node, traverse the tree of nodes under it.
     */
    function visit(n) {
        if (!tsm.Node.isCallExpression(n)) {
            n.forEachChild(visit);
            return;
        }
        const exp = n.getExpression();
        if (!tsm.Node.isPropertyAccessExpression(exp)) {
            n.forEachChild(visit);
            return;
        }
        const expText = exp.getExpression().getText();
        const propName = exp.getName();
        if (expText !== 'schema' || propName !== 'addToContext') {
            n.forEachChild(visit);
            return;
        }
        log.trace('found call', { text: n.getText() });
        // Get the argument passed to addToContext so we can extract its type
        if (n.getArguments().length === 0) {
            log.trace('no args passed to call, the user should see a static type error, stopping extraction');
            return;
        }
        if (n.getArguments().length > 1) {
            log.trace('multiple args passed to call, the user should see a static type error, stopping extraction');
            return;
        }
        const contextAdder = n.getArguments()[0];
        const contextAdderType = contextAdder.getType(); // checker.getTypeAtLocation(contextAdder)
        const contextAdderSigs = contextAdderType.getCallSignatures();
        log.trace('found call arg', { text: contextAdder.getText() });
        // Get the signature of the argument so we can extract its return type
        if (contextAdderSigs.length === 0) {
            log.trace('arg had no signatures, this means the user passed a non-callable, the user should see a static type error, stopping context type extraction');
            return;
        }
        if (contextAdderSigs.length > 1) {
            log.warn('An overloaded function passed to addToContext. The first signature will be taken. This choice is arbitrary and may result in bad context extraction.', { text: contextAdder.getText() });
        }
        const contextAdderSig = contextAdderSigs[0];
        const contextAdderRetType = unwrapMaybePromise(contextAdderSig.getReturnType());
        let contextAdderRetTypeString = contextAdderRetType.getText(undefined, tsm.ts.TypeFormatFlags.NoTruncation);
        if (contextAdderRetType.isUnion()) {
            const unionTypes = contextAdderRetType.getUnionTypes();
            // If every member of the union are objects or interfaces
            if (unionTypes.every((t) => t.isObject() || t.isInterface())) {
                contextAdderRetTypeString = mergeUnionTypes(unionTypes);
            }
            else {
                throw errors_1.forbiddenUnionTypeError({
                    unionType: contextAdderRetType.getText(undefined, tsm.ts.TypeFormatFlags.NoTruncation),
                });
            }
        }
        if (contextAdderRetType.isInterface() || contextAdderRetType.getAliasSymbol()) {
            const info = extractTypeImportInfoFromType(contextAdderRetType);
            if (info) {
                typeImportsIndex[info.name] = info;
            }
            contextTypeContributions.types.push(contribTypeRef(contextAdderRetTypeString));
            return;
        }
        contextTypeContributions.types.push(contribTypeLiteral(contextAdderRetTypeString));
        // search for named references, they will require importing later on
        const contextAdderRetProps = contextAdderRetType.getProperties();
        for (const prop of contextAdderRetProps) {
            log.trace('processing prop', { name: prop.getName() });
            const tsmn = prop.getDeclarations()[0];
            const t = tsmn.getType();
            if (t)
                if (t.getAliasSymbol()) {
                    log.trace('found alias', {
                        type: t.getText(undefined, typescript_1.default.TypeFormatFlags.NoTruncation),
                    });
                    const info = extractTypeImportInfoFromType(t);
                    if (info) {
                        typeImportsIndex[info.name] = info;
                    }
                }
                else if (t.isIntersection()) {
                    log.trace('found intersection', {
                        types: t.getIntersectionTypes().map((t) => t.getText(undefined, typescript_1.default.TypeFormatFlags.NoTruncation)),
                    });
                    const infos = t
                        .getIntersectionTypes()
                        .map((t) => extractTypeImportInfoFromType(t))
                        .filter((info) => info !== null);
                    if (infos.length) {
                        infos.forEach((info) => {
                            typeImportsIndex[info.name] = info;
                        });
                    }
                }
                else if (t.isUnion()) {
                    log.trace('found union', {
                        types: t.getUnionTypes().map((t) => t.getText(undefined, typescript_1.default.TypeFormatFlags.NoTruncation)),
                    });
                    const infos = t
                        .getUnionTypes()
                        .map((t) => extractTypeImportInfoFromType(t))
                        .filter((info) => info !== null);
                    if (infos.length) {
                        infos.forEach((info) => {
                            typeImportsIndex[info.name] = info;
                        });
                    }
                }
                else {
                    const info = extractTypeImportInfoFromType(t);
                    if (info) {
                        typeImportsIndex[info.name] = info;
                    }
                }
        }
    }
}
exports.extractContextTypes = extractContextTypes;
function extractTypeImportInfoFromType(t) {
    var _a;
    let sym = t.getAliasSymbol();
    let name = sym === null || sym === void 0 ? void 0 : sym.getName();
    log.trace('found prop type alias symbol?', { found: !!sym });
    if (t.isArray()) {
        return extractTypeImportInfoFromType(t.getArrayElementTypeOrThrow());
    }
    if (!sym) {
        sym = t.getSymbol();
        log.trace('found prop type symbol?', { found: !!sym });
        if (!sym)
            return null;
        name = sym.getName();
        // not alias but is inline, then skip
        if (name === '__object')
            return null;
        if (name === '__type')
            return null;
    }
    log.trace('found name?', { name });
    if (!name)
        return null;
    const d = (_a = sym.getDeclarations()) === null || _a === void 0 ? void 0 : _a[0];
    if (!d)
        throw utils_1.exception('A type with a symbol but the symbol has no declaration', {});
    const sourceFile = d.getSourceFile();
    const { modulePath, isNode } = getAbsoluteImportPath(sourceFile);
    return {
        name: name,
        modulePath: modulePath,
        isExported: sourceFile.getExportedDeclarations().has(name),
        isNode: isNode,
    };
}
function unwrapMaybePromise(type) {
    var _a;
    if (((_a = type.getSymbol()) === null || _a === void 0 ? void 0 : _a.getName()) === 'Promise') {
        const typeArgs = type.getTypeArguments();
        if (typeArgs.length > 0) {
            const wrappedType = typeArgs[0];
            return wrappedType;
        }
    }
    return type;
}
exports.unwrapMaybePromise = unwrapMaybePromise;
function getAbsoluteImportPath(sourceFile) {
    var _a, _b;
    let isNode = false;
    let modulePath = Path.join(Path.dirname(sourceFile.getFilePath()), sourceFile.getBaseNameWithoutExtension());
    const nodeModule = (_a = modulePath.match(/node_modules\/@types\/node\/(.+)/)) === null || _a === void 0 ? void 0 : _a[1];
    if (nodeModule) {
        modulePath = nodeModule;
        isNode = true;
    }
    else {
        const externalPackage = (_b = modulePath.match(/node_modules\/@types\/(.+)/)) === null || _b === void 0 ? void 0 : _b[1];
        if (externalPackage) {
            modulePath = externalPackage;
        }
    }
    return { isNode, modulePath };
}
function mergeUnionTypes(unionTypes) {
    const properties = unionTypes.reduce((acc, u) => {
        u.getProperties().forEach((p) => {
            const name = p.getName();
            const isOptional = p.hasFlags(tsm.ts.SymbolFlags.Optional);
            const propertyType = p
                .getDeclarations()[0]
                .getType()
                .getText(undefined, tsm.ts.TypeFormatFlags.NoTruncation);
            if (!acc[name]) {
                acc[name] = [];
            }
            acc[name].push({ isOptional, type: propertyType });
        });
        return acc;
    }, {});
    const stringifiedProps = Object.entries(properties)
        .map(([name, propertiesInfo]) => {
        // If a property is not present across all members of the union type, force it to be optional
        const isOptional = propertiesInfo.length !== unionTypes.length ? true : propertiesInfo.some((p) => p.isOptional);
        const typesOfProperty = lodash_1.default(propertiesInfo)
            .flatMap((p) => p.type.split(' | '))
            .uniq()
            .value()
            .join(' | ');
        return `${name}${isOptional ? '?' : ''}: ${typesOfProperty};`;
    })
        .join(' ');
    // merge the properties
    return `{ ${stringifiedProps} }`;
}
exports.mergeUnionTypes = mergeUnionTypes;
//# sourceMappingURL=extractor.js.map